from infrasys import BaseQuantity
import scipy.stats as stats


class ProbabilityFunctionBuilder:
    """Class containing utility fuctions for sceario definations."""

    def __init__(self, dist, params: list[float | BaseQuantity]):
        """Constructor for BaseScenario class.

        Args:
            dist (str): Name of teh distribution. Should follow Scipy naming convention
            params (list): A list of parameters for the chosen distribution function. See Scipy.stats documentation
        """
        base_quantity = [p for p in params if isinstance(p, BaseQuantity)][0]
        self.quantity = base_quantity.__class__
        self.units = base_quantity.units
        self.dist = getattr(stats, dist)
        self.params = [p.magnitude if isinstance(p, BaseQuantity) else p for p in params]
        return

    def sample(self):
        """Sample the distribution"""
        return self.quantity(self.dist.rvs(*self.params, size=1)[0], self.units)

    def probability(self, value: BaseQuantity) -> float:
        """Calculates survival probability of a given asset.

        Args:
            value (float): value for vetor of interest. Will change with scenarions
        """
        assert isinstance(value, BaseQuantity), "Value must be a BaseQuantity"

        cdf = self.dist.cdf
        try:
            return cdf(value.to(self.units).magnitude, *self.params)
        except Exception:
            return cdf(value, *self.params)



from datetime import datetime
import sqlite3
import os

from pydantic import field_serializer, field_validator
from gdm.quantities import Distance
from shapely.geometry import Point
import plotly.graph_objects as go
import pandas as pd

from erad.models.hazard.common import ERAD_DB, HISTROIC_EARTHQUAKE_TABLE
from erad.models.hazard.base_models import BaseDisasterModel


class EarthQuakeModel(BaseDisasterModel):
    timestamp: datetime
    origin: Point
    depth: Distance
    magnitude: float

    @field_validator("origin", mode="before")
    def deserialize_point(cls, value):
        if isinstance(value, dict) and value.get("type") == "Point":
            coords = value["coordinates"]
            return Point(coords)
        return value

    @field_serializer("origin")
    def serialize_location(self, point: Point, _info):
        return {"type": "Point", "coordinates": (point.x, point.y)}

    @classmethod
    def example(cls) -> "EarthQuakeModel":
        return EarthQuakeModel(
            name="earthquake 1",
            timestamp=datetime.now(),
            origin=Point(-120.93036, 36.60144),
            depth=Distance(300, "km"),
            magnitude=5.0,
        )

    @classmethod
    def from_earthquake_code(cls, earthquake_code: str) -> "EarthQuakeModel":
        assert os.path.exists(ERAD_DB), f"The data file {ERAD_DB} not found"

        conn = sqlite3.connect(ERAD_DB)
        earthquake_data = pd.read_sql(
            f"SELECT * FROM {HISTROIC_EARTHQUAKE_TABLE} WHERE ID = '{earthquake_code}'", conn
        )
        conn.close()

        assert not earthquake_data.empty, f"No earthquake {earthquake_code} found in the database"

        earthquake_data["Date"] = pd.to_datetime(earthquake_data["Date"])
        earthquake_data["Time"] = pd.to_datetime(earthquake_data["Time"])
        earthquake_data["DateTime"] = earthquake_data.apply(
            lambda row: datetime.combine(row["Date"].date(), row["Time"].time()), axis=1
        )

        long = earthquake_data.Longitude.values[0]
        lat = earthquake_data.Latitude.values[0]

        return cls(
            name=earthquake_code,
            timestamp=earthquake_data.DateTime.values[0].astype("datetime64[ms]").astype(datetime),
            origin=Point(long, lat),
            depth=Distance(earthquake_data.Depth.values[0], "km"),
            magnitude=earthquake_data.Magnitude.values[0],
        )

    def plot(
        self,
        time_index: int = 0,
        figure: go.Figure = go.Figure(),
        map_obj: type[go.Scattergeo | go.Scattermap] = go.Scattermap,
    ) -> int:
        figure.add_trace(
            map_obj(
                lat=[self.origin.y],
                lon=[self.origin.x],
                mode="markers",
                marker=dict(size=[self.magnitude * 10], color=[self.depth.magnitude], opacity=0.4),
                name="Earthquake",
                hovertext=[
                    f"""
                    <br> <b>Earthquake depth:</b> {self.depth}
                    <br> <b>Earthquake magnitude:</b> {self.magnitude}
                    """
                ],
                visible=(time_index == 0),
            )
        )
        return 1



from erad.models.hazard.base_models import BaseDisasterModel
from erad.models.hazard.earthquake import EarthQuakeModel
from erad.models.hazard.wild_fire import FireModel, FireModelArea
from erad.models.hazard.wind import WindModel
from erad.models.hazard.flood import FloodModel, FloodModelArea



from infrasys.base_quantity import BaseQuantity


class Speed(BaseQuantity):
    """Quantity representing speed."""

    __base_unit__ = "m/second"


class Acceleration(BaseQuantity):
    """Quantity representing acceleration."""

    __base_unit__ = "m/second**2"


class Temperature(BaseQuantity):
    """Quantity representing temperature."""

    __base_unit__ = "degC"


class Pressure(BaseQuantity):
    """Quantity representing pressure."""

    __base_unit__ = "millibar"


class Flow(BaseQuantity):
    """Quantity representing flow."""

    __base_unit__ = "feet**3/second"



from pathlib import Path
import json

from pydantic import BaseModel

from gdm.tracked_changes import TrackedChange


class EditStore(BaseModel):
    updates: list[TrackedChange] = []

    def to_json(self, filename: Path):
        filename = Path(filename)
        with open(filename, "w") as f:
            f.write(self.model_dump_json(indent=4))

    @classmethod
    def from_json(cls, filename: Path):
        filename = Path(filename)
        with open(filename) as f:
            data = json.load(f)
            return EditStore(**data)



from gdm.quantities import Distance

import erad.models.fragility_curve as frag
from erad.enums import AssetTypes

DEFAULT_FLOOD_DEPTH_FRAGILITY_CURVES = frag.HazardFragilityCurves(
    asset_state_param="flood_depth",
    curves=[
        frag.FragilityCurve(
            asset_type=AssetTypes.switch,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Distance(0.35, "m"), Distance(0.50, "m"), 1 / 0.35],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.battery_storage,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Distance(0.35, "m"), Distance(0.50, "m"), 1 / 0.35],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_junction_box,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Distance(0.40, "m"), Distance(1.0, "m"), 1 / 0.40],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_overhead_lines,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Distance(0.35, "m"), Distance(1.0, "m"), 1 / 0.35],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_poles,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Distance(0.35, "m"), Distance(1.0, "m"), 1 / 0.35],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_underground_cables,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Distance(0.8, "m"), Distance(1.0, "m"), 1 / 0.8],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.solar_panels,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Distance(0.35, "m"), Distance(0.6, "m"), 1 / 0.35],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.substation,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Distance(0.40, "m"), Distance(1.0, "m"), 1 / 0.4],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transformer_mad_mount,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Distance(0.35, "m"), Distance(0.6, "m"), 1 / 0.35],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transformer_pole_mount,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Distance(0.30, "m"), Distance(0.8, "m"), 1 / 0.3],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_junction_box,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Distance(0.40, "m"), Distance(1.0, "m"), 1 / 0.40],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_overhead_lines,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Distance(0.30, "m"), Distance(1.8, "m"), 1 / 0.3],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_tower,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Distance(0.40, "m"), Distance(2.2, "m"), 1 / 0.40],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_underground_cables,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Distance(0.25, "m"), Distance(0.8, "m"), 1 / 0.25],
            ),
        ),
    ],
)



import erad.models.fragility_curve as frag
from erad.enums import AssetTypes
from erad.quantities import Speed

DEFAULT_PEAK_GROUND_VELOCITY_FRAGILITY_CURVES = frag.HazardFragilityCurves(
    asset_state_param="peak_ground_velocity",
    curves=[
        frag.FragilityCurve(
            asset_type=AssetTypes.switch,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.5, "cm/s"), Speed(35, "cm/s"), 2]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.battery_storage,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.5, "cm/s"), Speed(35, "cm/s"), 2]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_junction_box,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.5, "cm/s"), Speed(35, "cm/s"), 2]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_overhead_lines,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.5, "cm/s"), Speed(40, "cm/s"), 2]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_poles,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.5, "cm/s"), Speed(40, "cm/s"), 2]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_underground_cables,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.5, "cm/s"), Speed(60, "cm/s"), 2]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.solar_panels,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.5, "cm/s"), Speed(35, "cm/s"), 2]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.substation,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.5, "cm/s"), Speed(50, "cm/s"), 2]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transformer_mad_mount,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.5, "cm/s"), Speed(35, "cm/s"), 2]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transformer_pole_mount,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.5, "cm/s"), Speed(40, "cm/s"), 2]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_junction_box,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.5, "cm/s"), Speed(50, "cm/s"), 2]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_overhead_lines,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.5, "cm/s"), Speed(45, "cm/s"), 2]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_tower,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.5, "cm/s"), Speed(35, "cm/s"), 2]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_underground_cables,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Speed(0.55, "cm/s"), Speed(65, "cm/s"), 1 / 0.55],
            ),
        ),
    ],
)



from pydantic import ValidationError
from infrasys import BaseQuantity

from erad.models.fragility_curve import HazardFragilityCurves, FragilityCurve, ProbabilityFunction
from erad.default_fragility_curves import DEFAULT_FRAGILTY_CURVES
from erad.quantities import Speed
from erad.enums import AssetTypes

import pytest


def test_valid_hazard_fragility_model():
    HazardFragilityCurves(
        asset_state_param="peak_ground_velocity",
        curves=[
            FragilityCurve(
                asset_type=AssetTypes.battery_storage,
                prob_function=ProbabilityFunction(
                    distribution="lognorm", parameters=[Speed(35, "cm/s"), 0.5]
                ),
            ),
        ],
    )


def test_invalid_hazard_fragility_model():
    with pytest.raises(ValidationError):
        HazardFragilityCurves(
            asset_state_param="invalid_asset_model_attribute",
            curves=[
                FragilityCurve(
                    asset_type=AssetTypes.battery_storage,
                    prob_function=ProbabilityFunction(
                        distribution="lognorm", parameters=[Speed(35, "cm/s"), 0.5]
                    ),
                ),
            ],
        )


@pytest.mark.parametrize("hazard_curve_set", DEFAULT_FRAGILTY_CURVES)
def test_default_hazard_curves(hazard_curve_set: HazardFragilityCurves):
    for curve in hazard_curve_set.curves:
        prob = curve.prob_function
        prob_model = prob.prob_model
        sample = prob_model.sample()
        assert isinstance(sample, BaseQuantity)
        probability = prob_model.probability(prob.prob_model.quantity(45, prob.prob_model.units))
        assert probability >= 0 and probability <= 1


def test_valid_fragility_curve():
    (ProbabilityFunction(distribution="lognorm", parameters=[Speed(35, "cm/s"), 0.5]),)


def test_invalid_fragility_curve():
    with pytest.raises(ValidationError):
        (ProbabilityFunction(distribution="not_valid_dist", parameters=[Speed(35, "cm/s"), 0.5]),)

    with pytest.raises(ValidationError):
        (ProbabilityFunction(distribution="lognorm", parameters=[35, 0.5]),)

    with pytest.raises(ValidationError):
        (
            ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(35, "cm/s"), Speed(35, "m/s")]
            ),
        )


def test_valid_probability_func_use():
    prob_data_model = ProbabilityFunction(
        distribution="lognorm", parameters=[Speed(35, "cm/s"), 0.5]
    )
    prob_model = prob_data_model.prob_model
    sample = prob_model.sample()
    assert isinstance(sample, BaseQuantity)
    probability = prob_model.probability(Speed(45, "cm/s"))
    assert probability >= 0 and probability <= 1


def test_invalid_probability_func_use():
    with pytest.raises(AssertionError):
        prob_data_model = ProbabilityFunction(
            distribution="lognorm", parameters=[Speed(35, "cm/s"), 0.5]
        )
        prob_model = prob_data_model.prob_model
        prob_model.probability(45)



from gdm.tracked_changes import filter_tracked_changes_by_name_and_date, apply_updates_to_system
from gdm.distribution import DistributionSystem

from erad.runner import HazardScenarioGenerator
from erad.systems.asset_system import AssetSystem
from erad.systems.hazard_system import HazardSystem


def test_hazard_scenarios(gdm_system: DistributionSystem):
    number_of_samples = 5
    asset_system = AssetSystem.from_gdm(gdm_system)
    hazard_system = HazardSystem.earthquake_example()
    scenario_generator = HazardScenarioGenerator(
        asset_system=asset_system, hazard_system=hazard_system
    )
    scenarios = scenario_generator.samples(number_of_samples)
    assert len(scenarios) == 91

    scenario_names = {s.scenario_name for s in scenarios}
    assert len(scenario_names) == number_of_samples

    scenario_0 = filter_tracked_changes_by_name_and_date(
        tracked_changes=scenarios, scenario_name=list(scenario_names)[0]
    )
    n_asset_inservice_before = sum(
        [c.in_service for c in gdm_system.iter_all_components() if hasattr(c, "in_service")]
    )
    updated_gdm_system = apply_updates_to_system(scenario_0, gdm_system, None)
    n_asset_inservice_after = sum(
        [
            c.in_service
            for c in updated_gdm_system.iter_all_components()
            if hasattr(c, "in_service")
        ]
    )
    assert n_asset_inservice_before != n_asset_inservice_after



from infrasys import Component

import pytest

from erad.systems.hazard_system import HazardSystem
from erad.models.hazard import WindModel
from erad.constants import HAZARD_MODELS


def test_component_addition():
    h = HazardSystem(auto_add_composed_components=True)
    for m in HAZARD_MODELS:
        h.add_component(m.example())


def test_component_failure():
    class Testing(Component):
        ...

    test = Testing(name="asdf")

    h = HazardSystem(auto_add_composed_components=True)
    with pytest.raises(AssertionError):
        h.add_component(test)


def test_system_serialization_deserialization(tmp_path):
    h = HazardSystem(auto_add_composed_components=True)
    for m in HAZARD_MODELS:
        h.add_component(m.example())
    h.to_json(tmp_path / "test_system.json")

    HazardSystem.from_json(tmp_path / "test_system.json")


def test_earthquake_example():
    HazardSystem.earthquake_example()


def test_fire_example():
    HazardSystem.fire_example()


def test_wind_example():
    HazardSystem.wind_example()


def test_flood_example():
    HazardSystem.flood_example()


def test_wind_plot():
    hazard = WindModel.from_hurricane_sid("2017228N14314")
    system = HazardSystem(auto_add_composed_components=True)
    system.add_components(*hazard)
    system.plot()


def test_earthquake_plot():
    system = HazardSystem.earthquake_example()
    system.plot()


def test_wildfire_plot():
    system = HazardSystem.fire_example()
    system.plot()


def test_flood_plot():
    system = HazardSystem.flood_example()
    system.plot()



from datetime import datetime

from gdm.distribution.components import DistributionBus
from gdm.distribution import DistributionSystem
from gdm.quantities import Distance
from shapely.geometry import Point

from erad.systems.hazard_system import HazardSystem
from erad.models.hazard import EarthQuakeModel
from erad.runner import HarzardSimulator


def test_gdm_model_earthquake(gdm_system: DistributionSystem):
    hazard_scenario = HarzardSimulator.from_gdm(gdm_system)
    buses: list[DistributionBus] = list(gdm_system.get_components(DistributionBus))
    buses = sorted(buses, key=lambda b: b.name)
    earthquake = EarthQuakeModel(
        name="earthquake_1",
        timestamp=datetime.now(),
        origin=Point(buses[0].coordinate.y, buses[0].coordinate.x),
        depth=Distance(100, "kilometer"),
        magnitude=5.8,
    )
    hazard_system = HazardSystem(auto_add_composed_components=True)
    hazard_system.add_component(earthquake)
    hazard_scenario.run(hazard_system)


def test_asset_graph_undirected(gdm_system: DistributionSystem):
    dist_graph = gdm_system.get_undirected_graph()
    hazard_scenario = HarzardSimulator.from_gdm(gdm_system)
    graph = hazard_scenario.asset_system.get_undirected_graph()
    assert (
        dist_graph.number_of_edges() == graph.number_of_edges()
    ), f"The number of edges in the asset graph ({graph.number_of_edges()}) should match the distribution system graph ({dist_graph.number_of_edges()})."
    assert (
        dist_graph.number_of_nodes() == graph.number_of_nodes()
    ), f"The number of nodes in the asset graph ({graph.number_of_nodes()}) should match the distribution system graph. ({dist_graph.number_of_nodes()})"


def test_asset_graph_directed(gdm_system: DistributionSystem):
    dist_graph = gdm_system.get_directed_graph()
    hazard_scenario = HarzardSimulator.from_gdm(gdm_system)
    graph = hazard_scenario.asset_system.get_dircted_graph()

    assert (
        dist_graph.number_of_edges() == graph.number_of_edges()
    ), f"The number of edges in the asset graph ({graph.number_of_edges()}) should match the distribution system graph ({dist_graph.number_of_edges()})."
    assert (
        dist_graph.number_of_nodes() == graph.number_of_nodes()
    ), f"The number of nodes in the asset graph ({graph.number_of_nodes()}) should match the distribution system graph. ({dist_graph.number_of_nodes()})"



from erad.models.hazard import EarthQuakeModel, FireModel, WindModel


def test_earthquake_from_history():
    EarthQuakeModel.from_earthquake_code("ISCGEM851547")


def test_fire_from_history():
    FireModel.from_wildfire_name("GREAT LAKES FIRE")


def test_hurricane_from_history():
    WindModel.from_hurricane_sid("2017228N14314")



from erad.default_fragility_curves import DEFAULT_FRAGILTY_CURVES


def test_plotting(tmp_path):
    """Test plotting of fragility curves."""

    for i, hazard_curves in enumerate(DEFAULT_FRAGILTY_CURVES):
        img = tmp_path / f"test_plotting_{i}.html"
        hazard_curves.plot(img, 0, 80, 1000)

        assert img.exists(), "Plotting failed, image not created."



from erad.models.asset import Asset
from erad.runner import HarzardSimulator
from erad.systems.asset_system import AssetSystem
from erad.systems.hazard_system import HazardSystem


def get_asset_system() -> AssetSystem:
    asset = Asset.example()
    asset_system = AssetSystem(auto_add_composed_components=True)
    asset_system.add_component(asset.example())
    return asset_system


def test_earthquake_simulation():
    hazard_scenario = HarzardSimulator(asset_system=get_asset_system())
    hazard_scenario.run(hazard_system=HazardSystem.earthquake_example())


def test_fire_simulation():
    hazard_scenario = HarzardSimulator(asset_system=get_asset_system())
    hazard_scenario.run(hazard_system=HazardSystem.fire_example())


def test_wind_simulation():
    hazard_scenario = HarzardSimulator(asset_system=get_asset_system())
    hazard_scenario.run(hazard_system=HazardSystem.wind_example())


def test_flood_simulation():
    hazard_scenario = HarzardSimulator(asset_system=get_asset_system())
    hazard_scenario.run(hazard_system=HazardSystem.flood_example())



from pathlib import Path

from loguru import logger
import requests


TEST_PATH = Path(__file__).parents[4] / "tests"
DATA_FOLDER_NAME = "data"
DATA_FOLDER = TEST_PATH / DATA_FOLDER_NAME
DB_FILENAME = "erad_data.sqlite"
ERAD_DB = TEST_PATH / DATA_FOLDER_NAME / DB_FILENAME
BUCKET_NAME = "erad_v2_dataset"

if not ERAD_DB.exists():
    logger.info("Erad database not found. Downloading from Google Cloud.")
    ERAD_DB.parent.mkdir(parents=True, exist_ok=True)
    url = f"https://storage.googleapis.com/{BUCKET_NAME}/{DB_FILENAME}"
    response = requests.get(url)

    with open(ERAD_DB, "wb") as f:
        f.write(response.content)
    logger.info("Download complete...")

HISTROIC_EARTHQUAKE_TABLE = "historic_earthquakes"
HISTROIC_HURRICANE_TABLE = "historic_hurricanes"
HISTROIC_FIRE_TABLE = "historic_fires"



import erad.models.fragility_curve as frag
from erad.quantities import Speed
from erad.enums import AssetTypes

DEFAULT_FLOOD_VELOCITY_FRAGILITY_CURVES = frag.HazardFragilityCurves(
    asset_state_param="flood_velocity",
    curves=[
        frag.FragilityCurve(
            asset_type=AssetTypes.switch,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Speed(0.35, "m/s"), Speed(1.5, "m/s"), 1 / 0.35],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.battery_storage,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Speed(0.35, "m/s"), Speed(1.5, "m/s"), 1 / 0.35],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_junction_box,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.4, "m/s"), Speed(2.0, "m/s"), 1 / 0.40]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_overhead_lines,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Speed(0.35, "m/s"), Speed(3.5, "m/s"), 1 / 0.35],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_poles,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Speed(0.35, "m/s"), Speed(3.0, "m/s"), 1 / 0.35],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.distribution_underground_cables,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.4, "m/s"), Speed(2.0, "m/s"), 1 / 0.4]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.solar_panels,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Speed(0.35, "m/s"), Speed(2.0, "m/s"), 1 / 0.35],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.substation,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.4, "m/s"), Speed(2.0, "m/s"), 1 / 0.4]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transformer_mad_mount,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Speed(0.35, "m/s"), Speed(1.5, "m/s"), 1 / 0.35],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transformer_pole_mount,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.3, "m/s"), Speed(2.5, "m/s"), 1 / 0.30]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_junction_box,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.4, "m/s"), Speed(1.5, "m/s"), 1 / 0.4]
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_overhead_lines,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Speed(0.35, "m/s"), Speed(3.5, "m/s"), 1 / 0.35],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_tower,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm",
                parameters=[Speed(0.40, "m/s"), Speed(3.0, "m/s"), 1 / 0.40],
            ),
        ),
        frag.FragilityCurve(
            asset_type=AssetTypes.transmission_underground_cables,
            prob_function=frag.ProbabilityFunction(
                distribution="lognorm", parameters=[Speed(0.40, "m/s"), Speed(2.0, "m/s"), 1 / 0.4]
            ),
        ),
    ],
)



from datetime import datetime
import sqlite3
import os

from shapely.geometry import MultiPolygon, Polygon, Point
from pydantic import field_serializer, field_validator
import plotly.graph_objects as go
from gdm.quantities import Angle
from infrasys import Component
from shapely import wkb
import pandas as pd

from erad.models.hazard.common import ERAD_DB, HISTROIC_FIRE_TABLE
from erad.models.hazard.base_models import BaseDisasterModel
from erad.quantities import Speed


class FireModelArea(Component):
    name: str = ""
    affected_area: Polygon
    wind_speed: Speed
    wind_direction: Angle

    @field_validator("affected_area", mode="before")
    def deserialize_polygon(cls, value):
        if isinstance(value, dict) and value.get("type") == "Polygon":
            points = [Point(c) for c in value["coordinates"]]
            return Polygon(points)
        return value

    @field_serializer("affected_area")
    def serialize_polygon(self, poly: Polygon, _info):
        return {"type": "Polygon", "coordinates": list(poly.exterior.coords)}

    @classmethod
    def example(cls) -> "FireModelArea":
        return FireModelArea(
            affected_area=Polygon(
                [
                    (-120.93036, 36.60144),
                    (-120.91072, 36.60206),
                    (-120.91127, 36.5712),
                    (-120.93405, 36.58100),
                ]
            ),
            wind_speed=Speed(50, "miles/hour"),
            wind_direction=Angle(45, "deg"),
        )


class FireModel(BaseDisasterModel):
    timestamp: datetime
    affected_areas: list[FireModelArea]

    @classmethod
    def example(cls) -> "FireModel":
        return FireModel(
            name="fire 1",
            timestamp=datetime.now(),
            affected_areas=[FireModelArea.example()],
        )

    @classmethod
    def from_wildfire_name(cls, wildfire_name: str) -> "FireModel":
        assert os.path.exists(ERAD_DB), f"The data file {ERAD_DB} not found"
        conn = sqlite3.connect(ERAD_DB)
        fire_data = pd.read_sql(
            f"SELECT * FROM {HISTROIC_FIRE_TABLE} WHERE firename = '{wildfire_name}';", conn
        )
        if fire_data.empty:
            raise ValueError(
                f"Fire '{wildfire_name}'  not found in  table '{HISTROIC_FIRE_TABLE}' in the database"
            )
        conn.close()
        fire_data["discoverydatetime"] = pd.to_datetime(fire_data["discoverydatetime"])
        geometry: MultiPolygon = [wkb.loads(g) for g in fire_data.GEOMETRY][0]
        areas = []
        for i, poly in enumerate(geometry.geoms):
            areas.append(
                FireModelArea(
                    affected_area=poly,
                    wind_speed=Speed(-999, "miles/hour"),
                    wind_direction=Angle(0, "deg"),
                )
            )
        return cls(
            name=wildfire_name,
            timestamp=fire_data["discoverydatetime"]
            .values[0]
            .astype("datetime64[ms]")
            .astype(datetime),
            affected_areas=areas,
        )

    def plot(
        self,
        time_index: int = 0,
        figure: go.Figure = go.Figure(),
        map_obj: type[go.Scattergeo | go.Scattermap] = go.Scattermap,
    ) -> int:
        for area in self.affected_areas:
            lon, lat = area.affected_area.exterior.xy  # returns x and y sequences
            figure.add_trace(
                map_obj(
                    lon=lon.tolist(),
                    lat=lat.tolist(),
                    mode="markers+lines+text",
                    fill="toself",
                    marker={
                        "size": 10,
                        "color": [area.wind_speed.magnitude],
                    },
                    hovertemplate=f"""
                        <br> <b>Wind speed:</b> {area.wind_speed}
                        <br> <b>Wind direction:</b> {area.wind_direction}
                        """,
                    hoverinfo="text",
                    visible=(time_index == 0),
                )
            )
        return len(self.affected_areas)



from typing import Type, Callable

from pydantic import BaseModel, ConfigDict
from infrasys import Component

from erad.enums import AssetTypes


class BaseEradScenarioModel(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")


class ComponentFilterModel(BaseEradScenarioModel):
    component_type: Type[Component]
    component_filter: Callable | None = None


class AssetComponentMap(BaseEradScenarioModel):
    asset_type: AssetTypes
    filters: list[ComponentFilterModel] = []



from datetime import datetime

from pydantic import field_serializer, field_validator
from shapely.geometry import Polygon, Point
from gdm.quantities import Distance
import plotly.graph_objects as go
from infrasys import Component

from erad.models.hazard.base_models import BaseDisasterModel
from erad.quantities import Speed


class FloodModelArea(Component):
    name: str = ""
    affected_area: Polygon
    water_velocity: Speed
    water_elevation: Distance

    @field_validator("affected_area", mode="before")
    def deserialize_polygon(cls, value):
        if isinstance(value, dict) and value.get("type") == "Polygon":
            points = [Point(c) for c in value["coordinates"]]
            return Polygon(points)
        return value

    @field_serializer("affected_area")
    def serialize_polygon(self, poly: Polygon, _info):
        return {"type": "Polygon", "coordinates": list(poly.exterior.coords)}

    @classmethod
    def example(cls) -> "FloodModelArea":
        return FloodModelArea(
            affected_area=Polygon(
                [
                    (-120.93036, 36.60144),
                    (-120.91072, 36.60206),
                    (-120.91127, 36.5712),
                    (-120.93405, 36.58100),
                ]
            ),
            water_velocity=Speed(50, "meter/second"),
            water_elevation=Distance(10, "feet"),
        )


class FloodModel(BaseDisasterModel):
    timestamp: datetime
    affected_areas: list[FloodModelArea]

    @classmethod
    def example(cls) -> "FloodModel":
        return FloodModel(
            name="flood 1",
            timestamp=datetime.now(),
            affected_areas=[FloodModelArea.example()],
        )

    def plot(
        self,
        time_index: int = 0,
        figure: go.Figure = go.Figure(),
        map_obj: type[go.Scattergeo | go.Scattermap] = go.Scattermap,
    ) -> int:
        for area in self.affected_areas:
            lon, lat = area.affected_area.exterior.xy  # returns x and y sequences
            figure.add_trace(
                map_obj(
                    lon=lon.tolist(),
                    lat=lat.tolist(),
                    mode="markers+lines+text",
                    fill="toself",
                    marker={
                        "size": 10,
                        "color": [area.water_elevation.magnitude],
                    },
                    hovertemplate=f"""
                        <br> <b>Water velocity:</b> {area.water_velocity}
                        <br> <b>Water elevation:</b> {area.water_elevation}
                        """,
                    visible=(time_index == 0),
                )
            )
        return len(self.affected_areas)
